#  MIT License Copyright (c) 2020. Houfu Ang
import logging
from typing import List

from pdpc_decisions.download_file import check_pdf
from .scraper import PDPCDecisionItem


def get_enforcement(items: List[PDPCDecisionItem]) -> None:
    """
    Get enforcement details for a list of PDPCDecisionItem.
    The results are saved as the field `enforcement` in the item in the following format.
    (**Type of enforcement**, **the penalty sum for financial penalties**)
    :param items: A list of PDPCDecisionItems to be assessed.
    :return: None
    """
    import spacy
    from spacy.matcher import Matcher
    nlp = spacy.load('en_core_web_sm')
    matcher = Matcher(nlp.vocab)
    financial_penalty_pattern = [{'LOWER': 'financial'},
                                 {'LOWER': 'penalty'},
                                 {'POS': 'ADP'},
                                 {'LOWER': '$'},
                                 {'LIKE_NUM': True}]
    financial_1_id = 'financial_1'
    matcher.add(financial_1_id, [financial_penalty_pattern])
    financial_penalty_pattern2 = [{'LOWER': 'financial'},
                                  {'LOWER': 'penalty'},
                                  {'POS': 'ADP'},
                                  {'LOWER': '$'},
                                  {'LIKE_NUM': True},
                                  {'LOWER': 'and'},
                                  {'LOWER': '$'},
                                  {'LIKE_NUM': True}]
    financial_2_id = 'financial_2'
    matcher.add(financial_2_id, [financial_penalty_pattern2])
    warning_pattern = [{'LOWER': 'warning'},
                       {'POS': 'AUX'},
                       {'LOWER': 'issued'}]
    warning_id = 'warning'
    matcher.add(warning_id, [warning_pattern])
    directions_pattern = [{'LOWER': 'directions'},
                          {'POS': 'AUX'},
                          {'LOWER': 'issued'}]
    directions_id = 'directions'
    matcher.add(directions_id, [directions_pattern])
    logging.info('Adding enforcement information to items.')
    for item in items:
        doc = nlp(item.summary)
        matches = matcher(doc)
        item.enforcement = []
        for match in matches:
            match_id, _, end = match
            if nlp.vocab.strings[financial_2_id] in match:
                span1 = doc[end - 4: end - 3]
                value = ['financial', int(span1.text.replace(',', ''))]
                if not item.enforcement.count(value):
                    item.enforcement.append(value)
                span2 = doc[end - 1:end]
                value = ['financial', int(span2.text.replace(',', ''))]
                if not item.enforcement.count(value):
                    item.enforcement.append(value)
            if nlp.vocab.strings[financial_1_id] in match:
                span = doc[end - 1:end]
                value = ['financial', int(span.text.replace(',', ''))]
                if not item.enforcement.count(value):
                    item.enforcement.append(value)
            if nlp.vocab.strings[warning_id] in match:
                item.enforcement.append(warning_id)
            if nlp.vocab.strings[directions_id] in match:
                item.enforcement.append(directions_id)


def get_decision_citation_all(items: List[PDPCDecisionItem]) -> None:
    """
    Gets the citation and case number for a list of PDPCDecisionItem.
    The result is saved in fields "citation" and "case_number" respectively.
    :param items:
    :return: None
    """
    logging.info('Adding citation information to items.')
    for item in items:
        item.citation, item.case_number = get_decision_citation_item(item)


def get_decision_citation_item(source: PDPCDecisionItem) -> (str, str):
    """
    Gets the citation and case number for a PDPCDecisionItem.
    :param source: The PDPCDecisionItem to get the citation and case number.
    :return: A tuple consisting of (citation, case_number)
    """
    from pdfminer.high_level import extract_text_to_fp
    import requests
    import io
    import re
    r = requests.get(source.download_url)
    citation = ''
    case_number = ''
    if check_pdf(source.download_url):
        with io.BytesIO(r.content) as pdf, io.StringIO() as output_string:
            extract_text_to_fp(pdf, output_string, page_numbers=[0, 1])
            contents = output_string.getvalue()
        summary_match = re.search(r'SUMMARY OF THE DECISION', contents)
        if not summary_match:
            citation_match = re.search(r'(\[\d{4}])\s+((?:\d\s+)?[A-Z|()]+)\s+\[?(\d+)\]?', contents)
            if citation_match:
                citation = citation_match.expand(r'\1 \2 \3')
        case_match = re.search(r'DP-\w*-\w*', contents)
        if case_match:
            case_number = case_match.group()
    return citation, case_number


def get_case_references_all(items: List[PDPCDecisionItem]) -> None:
    """
    Gets the references to cases in each PDPCDecisionItem in items.
    Adds the PDPCDecisionItem to a list of references in a target PDPCDecisionItem.
    The actions are saved in a parameter "referring_to" and "referred_by" of a PDPCDecisionItem respectively.
    :param items:
    :return:
    """
    logging.info('Adding case reference information to items.')
    # construct referring to index
    for item in items:
        citation, _ = get_decision_citation_item(source=item)
        item.referring_to = get_referring_to_item(item, citation=citation)
        add_referred_by_to_item(source=item, target=items, citation=citation)


def add_referred_by_to_item(source: PDPCDecisionItem, target: List[PDPCDecisionItem], citation: str) -> None:
    """
    Adds the source PDPCDecisionItem to a PDPCDecisionItem's "referred_by" field in target
    if source makes a reference to an item in target. The method finds references by citation.
    :param citation: Citation of source.
    :param source:
    :param target:
    :return:
    """
    if not hasattr(source, 'referring_to'):
        return
    for reference in source.referring_to:
        result_item = next((x for x in target if x.citation == reference), None)
        if result_item:
            if not hasattr(result_item, 'referred_by'):
                result_item.referred_by = [citation]
            if result_item.referred_by.count(citation) == 0:
                result_item.referred_by.append(citation)


def get_referring_to_item(source: PDPCDecisionItem, citation: str) -> List[str]:
    """
    Gets the references to cases made in source PDPCDecisionItem.
    :param citation: Citation of source
    :param source:
    :return: A list of case references/citations.
    """
    referring_to = []
    if check_pdf(source.download_url):
        from .download_file import get_text_from_pdf
        contents = get_text_from_pdf(source)
        import re
        citation_regex = re.compile(r'(\[\d{4}])\s+((?:\d\s+)?[A-Z|()]+)\s+\[?(\d+)\]?')
        citation_matches = citation_regex.finditer(contents)
        for match in citation_matches:
            result_citation = match.expand(r'\1 \2 \3')
            if (referring_to.count(result_citation) == 0) and (result_citation != citation):
                referring_to.append(result_citation)
    return referring_to


def scraper_extras(items):
    logging.info('Start adding extra information to items.')
    get_decision_citation_all(items)
    get_enforcement(items)
    get_case_references_all(items)
    logging.info('End adding extra information to items.')
    return True



#  MIT License Copyright (c) 2020. Houfu Ang

"""
Looks over the PDPC website and creates PDPC Decision objects

Requirements:
* Chrome Webdriver to automate web browser
"""
import logging
import re
from dataclasses import dataclass
from datetime import datetime
from typing import List

from selenium.webdriver import Chrome
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.remote.webelement import WebElement


def get_url(item: WebElement) -> str:
    """Gets the URL for the text of the decision."""
    link = item.find_element_by_tag_name('a')
    return link.get_property('href')


def get_summary(item: WebElement) -> str:
    """Gets the summary of a decision as provided by the PDPC."""
    return item.find_element_by_class_name('rte').text.split('\n')[0]


def get_published_date(item: WebElement) -> datetime.date:
    """Gets the date when the decision is published on the PDPC Website"""
    return datetime.strptime(item.find_element_by_class_name('page-date').text, "%d %b %Y").date()


def get_respondent(item: WebElement) -> str:
    """Gets the name of the respondent in the decision from title of the decision."""
    link = item.find_element_by_tag_name('h2')
    text = link.text
    return re.split(r"\s+[bB]y|[Aa]gainst\s+", text, re.I)[1].strip()


def get_title(item: WebElement) -> str:
    """Gets the title of the decision as provided by the PDPC"""
    return item.find_element_by_class_name('page-title').text


@dataclass
class PDPCDecisionItem:
    published_date: datetime.date
    respondent: str
    title: str
    summary: str
    download_url: str

    @classmethod
    def from_element(cls, decision: WebElement):
        """
        Create a PDPCDecisionItem from a section in the PDPC Website's list of commission's decisions.
        :param decision:
        :return:
        """
        published_date = get_published_date(decision)
        respondent = get_respondent(decision)
        title = get_title(decision)
        summary = get_summary(decision)
        download_url = get_url(decision)
        return cls(published_date, respondent, title, summary, download_url)

    def __str__(self):
        return "PDPCDecision object: {} {}".format(self.published_date, self.respondent)


class Scraper:
    def __init__(self):
        options = Options()
        options.add_argument('--headless')
        options.add_argument('--disable-gpu')
        options.add_argument('--window-size=1920,1080')
        options.add_argument('--no-sandbox')

        self.driver = Chrome(options=options)
        self.driver.implicitly_wait(25)

    @classmethod
    def scrape(cls,
               site_url="https://www.pdpc.gov.sg/All-Commissions-Decisions?"
                        "keyword=&industry=all&nature=all&decision=all&penalty=all&page=1") -> List[PDPCDecisionItem]:
        """Convenience method for scraping the PDPC Decision website with Scraper."""
        logging.info('Starting the scrape')
        self = cls()
        result = []
        try:
            self.driver.get(site_url)
            finished = False
            page_count = 1
            while not finished:
                items = self.driver.find_element_by_class_name('listing__list').find_elements_by_tag_name('li')
                for current_item in range(0, len(items)):
                    items = self.driver.find_element_by_class_name('listing__list').find_elements_by_tag_name('li')
                    from selenium.common.exceptions import NoSuchElementException
                    try:
                        link = items[current_item].find_element_by_tag_name('a').get_property('href')
                        self.driver.get(link)
                        decision = self.driver.find_element_by_class_name('detail-content')
                        item = PDPCDecisionItem.from_element(decision)
                        logging.info('Added: {}, {}'.format(item.respondent, item.published_date))
                        result.append(item)
                        self.driver.back()
                    except NoSuchElementException:
                        logging.warning("'detail-content' was not found: {}".format(self.driver.current_url))
                next_page = self.driver.find_element_by_class_name('pagination-next')
                if 'disabled' in next_page.get_attribute('class'):
                    finished = True
                else:
                    page_count += 1
                    new_url = "https://www.pdpc.gov.sg/All-Commissions-Decisions?page={}".format(page_count)
                    self.driver.get(new_url)
        finally:
            self.driver.close()
        logging.info('Ending scrape.')
        return result



#  MIT License Copyright (c) 2020. Houfu Ang

import io
import os
import re

import requests

from .scraper import PDPCDecisionItem


def download_files(options, items):
    print('Start downloading files.')
    if not os.path.exists(options["download_folder"]):
        os.mkdir(options["download_folder"])
    for item in items:
        print("Downloading a File: ", item.download_url)
        print("Date of Decision: ", item.published_date)
        print("Respondent: ", item.respondent)
        if check_pdf(item.download_url):
            download_pdf(options["download_folder"], item)
        else:
            download_text(options["download_folder"], item)
    print('Finished downloading files to ', options["download_folder"])


def check_pdf(download_url: str) -> bool:
    """Check if the download_url is a PDF or not by reading the extension."""
    from urllib.parse import urlparse
    result = urlparse(download_url)
    return result.path[-3:] == 'pdf'


def download_pdf(download_folder, item):
    destination_filename = "{} {}.pdf".format(item.published_date.strftime('%Y-%m-%d'), item.respondent)
    destination = os.path.join(download_folder, destination_filename)
    file_num = 0
    while os.path.isfile(destination):
        file_num += 1
        destination_filename = f"{item.published_date.strftime('%Y-%m-%d')} {item.respondent} ({file_num}).pdf"
        destination = os.path.join(download_folder, destination_filename)
    with open(destination, 'wb') as file:
        pdf_file = requests.get(item.download_url).content
        file.write(pdf_file)
    print("Downloaded a pdf: ", destination)
    return destination


def download_text(download_folder, item):
    destination_filename = "{} {}.txt".format(item.published_date.strftime('%Y-%m-%d'),
                                              item.respondent)
    destination = os.path.join(download_folder, destination_filename)
    with open(destination, "w", encoding='utf-8') as f:
        f.writelines(get_text_stream(item))
    print("Downloaded a text: ", destination)
    return destination


def get_text_from_item(source: PDPCDecisionItem) -> str:
    """Gets the text of the decision from the source item."""
    if check_pdf(source.download_url):
        return clean_up_source(get_text_from_pdf(source))
    else:
        return get_text_stream(source)


def get_text_from_pdf(source: PDPCDecisionItem) -> str:
    """
    Gets text from the decision of source in PDF. Avoids the first page if it is a cover page.
    :param source:
    :return:
    """
    r = requests.get(source.download_url)
    from pdfminer.high_level import extract_text_to_fp
    from pdfminer.layout import LAParams
    with io.BytesIO(r.content) as pdf, io.StringIO() as output_string, io.StringIO() as test_string:
        params = LAParams(line_margin=2)
        extract_text_to_fp(pdf, test_string, page_numbers=[0])
        first_page = test_string.getvalue()
        if len(first_page.split()) > 100:
            extract_text_to_fp(pdf, output_string, laparams=params)
        else:
            from pdfminer.pdfpage import PDFPage
            pages = len([page for page in PDFPage.get_pages(pdf)])
            extract_text_to_fp(pdf, output_string, page_numbers=[i for i in range(pages)[1:]], laparams=params)
        return output_string.getvalue()


def get_text_stream(item):
    from bs4 import BeautifulSoup
    soup = BeautifulSoup(requests.get(item.download_url).text, features='html5lib')
    rte = soup.find('div', class_='rte')
    text = rte.get_text()
    assert text, 'Download_text failed to get text from web page.'
    return text


def remove_extra_linebreaks(source):
    return [x for x in source if x != '' and not re.search(r'^\s+$', x)]


def remove_numbers_as_first_characters(source):
    return [x for x in source if not re.search(r'^\d*[\s.]*$', x)]


def remove_citations(source):
    return [x for x in source if not re.search(r'^\s+\[\d{4}]\s+(?:\d\s+)?[A-Z|()]+\s+\d+[\s.]?\s+$', x)]


def remove_feed_carriage(source):
    # identifies repeated instances (likely to be headers or footers)
    matches = [x for x in source if re.search(r'^\f', x)]
    counts = []
    for match in [ele for ind, ele in enumerate(matches, 1) if ele not in matches[ind:]]:
        counts.append((match, matches.count(match)))
    counts.sort(key=lambda m: m[1], reverse=True)
    if len(counts) > 1 and counts[0][1] > 1:
        return [x.replace('\f', '') for x in source if x != '\f' and x != counts[0][0]]
    else:
        return [x.replace('\f', '') for x in source if x != '\f']


def join_sentences_in_paragraph(source):
    result = []
    paragraph_string = ''
    for x in source:
        if re.search(r'\.\s*$', x):
            paragraph_string += x
            result.append(paragraph_string)
            paragraph_string = ''
        else:
            paragraph_string += x
    if paragraph_string != '':
        result.append(paragraph_string)
    return result


def clean_up_source(text):
    text_lines = text.split('\n')
    start_count = len(text_lines)
    text_lines = remove_extra_linebreaks(text_lines)
    text_lines = remove_numbers_as_first_characters(text_lines)
    text_lines = remove_citations(text_lines)
    text_lines = remove_feed_carriage(text_lines)
    text_lines = join_sentences_in_paragraph(text_lines)
    end_count = len(text_lines)
    reduced = (start_count - end_count) / start_count * 100
    print('Reduced from {} lines to {} lines. {:0.2f}% Wow'.format(start_count, end_count, reduced))
    return '\n'.join(text_lines)


def download_corpus_file(options, item):
    print("Source File: ", item.download_url)
    print("Date of Decision: ", item.published_date)
    print("Respondent: ", item.respondent)
    destination_filename = f"{item.published_date.strftime('%Y-%m-%d')} {item.respondent}.txt"
    destination = os.path.join(options["corpus_folder"], destination_filename)
    file_num = 0
    while os.path.isfile(destination):
        file_num += 1
        destination_filename = f"{item.published_date.strftime('%Y-%m-%d')} {item.respondent} ({file_num}).txt"
        destination = os.path.join(options["corpus_folder"], destination_filename)
    with open(destination, 'w') as fOut:
        text = get_text_from_item(item)
        fOut.write(text)
    print("Wrote: {}".format(destination))
    return destination


def create_corpus(options, items):
    print('Now creating corpus.')
    if not os.path.exists(options["corpus_folder"]):
        os.mkdir(options["corpus_folder"])
    for item in items:
        download_corpus_file(options, item)
    print('Number of items in corpus: ', len(items))
    print('Finished creating corpus at ', options["corpus_folder"])



#  MIT License Copyright (c) 2020. Houfu Ang

import csv


def save_scrape_results_to_csv(options, scrape_results):
    print('Saving scrape results as a csv file.')
    with open(options["csv_path"], 'w', newline='', encoding='utf-8') as f:
        csvwriter = csv.writer(f)
        if options['extras']:
            csvwriter.writerow(
                ['published_date', 'title', 'respondent', 'summary',
                 'download_url', 'citation', 'case_number',
                 'referred_by', 'referring_to', 'enforcement'])
            for result in scrape_results:
                csvwriter.writerow(
                    [result.published_date, result.title, result.respondent, result.summary,
                     result.download_url, result.citation, result.case_number,
                     result.referred_by, result.referring_to, result.enforcement])
        else:
            csvwriter.writerow(['published_date', 'title', 'respondent', 'summary', 'download_url'])
            for result in scrape_results:
                csvwriter.writerow(
                    [result.published_date, result.title, result.respondent, result.summary, result.download_url])
    print('Save completed, files saved at ', options["csv_path"])
